#pragma once

#include <chrono>
#include <event.h>
#include <deque>
#include <string>

#include <pipes/ssl.h>
#include <sys/queue.h>

namespace ts::server {
    class QueryClient;
}

namespace ts::server::server::query {
    enum struct ConnectionType {
        UNKNOWN,

        PLAIN_TEXT,
        SSL_ENCRYPTED,

        /* SSH */
    };

    enum struct ConnectionState {
        INITIALIZING,
        CONNECTED,
        DISCONNECTING,
        DISCONNECTED
    };

    enum struct CommandAssembleState {
        SUCCESS,
        MORE_COMMANDS_PENDING,

        NO_COMMAND_PENDING
    };

    class QueryClientConnection {
        public:
            explicit QueryClientConnection(QueryClient* /* client */, int /* file descriptor */);
            ~QueryClientConnection();

            [[nodiscard]] inline ConnectionType connection_type() const { return this->connection_type_; }

            bool initialize(std::string& /* error */);
            void add_read_event();

            void finalize(bool /* is destructor call */);

            void send_data(const std::string_view& /* payload */);
            void send_data_raw(const std::string_view& /* payload */);

            void enforce_text_connection();
            [[nodiscard]] CommandAssembleState next_command(std::string& /* command */);

            /* could be called from every thread (event IO thread) */
            void close_connection(const std::chrono::system_clock::time_point& /* disconnect timeout */);
        private:
            struct WriteBuffer {
                char* original_ptr;
                char* ptr;
                size_t length;

                TAILQ_ENTRY(WriteBuffer) tq;
            };

            QueryClient* client_handle{nullptr};
            ConnectionState connection_state{ConnectionState::INITIALIZING};
            std::chrono::system_clock::time_point disconnect_timeout{};

            ConnectionType connection_type_{ConnectionType::UNKNOWN};
            int file_descriptor_{-1};

            /* only delete the events within the event loop! */
            std::mutex event_mutex{};
            ::event* readEvent{nullptr};
            ::event* writeEvent{nullptr};

            pipes::SSL ssl_handler{};

            std::mutex buffer_lock{};
            struct {
                void* buffer{nullptr};
                size_t length{0};
                size_t fill_count{0};
                std::chrono::system_clock::time_point last_shrink{};
            } read_buffer;
            TAILQ_HEAD(, WriteBuffer) write_queue{};

            void handle_event_write(int, short);
            void handle_event_read(int, short);


            bool initialize_ssl();
            void handle_decoded_message(const void* /* message */, size_t /* length */);
    };
}