#pragma once

#include <tommath.h>
#include <memory>
#include <vector>
#include <misc/spin_lock.h>
#include <random>

namespace ts::server::server::udp {
    struct Puzzle {
        mp_int x;
        mp_int n;
        int level;

        mp_int result;

        uint8_t data_x[64];
        uint8_t data_n[64];
        uint8_t data_result[64];
    };

    class PuzzleManager {
        public:
            PuzzleManager();
            ~PuzzleManager();

            [[nodiscard]] bool precompute_puzzles(size_t amount);

            [[nodiscard]] size_t precomputed_puzzle_count();

            [[nodiscard]] std::shared_ptr<Puzzle> next_puzzle();
        private:
            void generate_puzzle(std::mt19937&);

            size_t cache_index{0};
            spin_lock cache_lock{};
            std::vector<std::shared_ptr<Puzzle>> cached_puzzles{};
    };
}