#pragma once

#include <string>
#include <memory>
#include <mutex>
#include <shared_mutex>
#include <Definitions.h>
#include <lock/rw_mutex.h>

namespace ts {
    class ServerChannel;
}

namespace ts::server {
    class ConnectedClient;

    namespace vserver {
        class VirtualServerBase;
    }
}

namespace ts::server::channels {
    enum struct ClientMoveResult {
        SUCCESS,
        TREE_LOCK_FAILED,
        VIEW_INSERT_FAILED
    };

    enum struct ChannelDeleteResult {
        SUCCESS,
        TREE_LOCK_FAILED,
        CLIENT_MOVE_FAILED
    };

    struct ChannelGroupInheritance {
        GroupId group_id{};
        ChannelId inherited_channel_id{};
    };

    class ClientChannelService {
        public:
            explicit ClientChannelService(vserver::VirtualServerBase*);
            ~ClientChannelService();

            bool initialize(std::string& /* error */);

            /* channel group related stuff */
            /* this will lock the channel tree in read mode */
            [[nodiscard]] ChannelGroupInheritance calculate_channel_group(ChannelId /* channel id */, ClientDbId /* client db id */, ClientType /* fallback type */);

            /*
             * ATTENTION: The methods client_ban, client_kick, client_move unlock the channel tree completely.
             *            All channel tree related variable must be checked again before using.
             */
            /* Note: Use only this method to disconnect the client and notify everybody else that he has been banned! */
            /* remove a client from the channel tree because he's banned */
            [[nodiscard]] ClientMoveResult client_ban(
                    const std::shared_ptr<ConnectedClient>& /* client */,
                    const std::shared_ptr<ConnectedClient>& /* invoker */,
                    const std::string& /* reason */,
                    size_t /* length */,
                    ts::rwshared_lock<ts::rw_mutex>& /* tree lock */);

            /* remove/move a client from/within the channel tree because he has been kicked */
            [[nodiscard]] ClientMoveResult client_kick(
                    const std::shared_ptr<ConnectedClient>& /* client */,
                    const std::shared_ptr<ConnectedClient>& /* invoker */,
                    const std::string& /* reason */,
                    const std::shared_ptr<ServerChannel>& /* target channel */,
                    ts::rwshared_lock<ts::rw_mutex>& /* tree lock */
            );

            /* move a client within the channel tree */
            [[nodiscard]] ClientMoveResult client_move(
                    const std::shared_ptr<ConnectedClient>& /* client */,
                    std::shared_ptr<ServerChannel> /* target channel */,
                    const std::shared_ptr<ConnectedClient>& /* invoker */,
                    const std::string& /* reason */,
                    ViewReasonId /* reason id */,
                    bool /* notify the client */,
                    ts::rwshared_lock<ts::rw_mutex>& /* tree lock */
            );

            [[nodiscard]] ChannelDeleteResult delete_channel(
                    std::shared_ptr<ServerChannel> /* target channel */,
                    const std::shared_ptr<ConnectedClient>& /* invoker */,
                    const std::string& /* kick message */,
                    ts::rwshared_lock<ts::rw_mutex>& /* tree lock */
            );
        private:
            vserver::VirtualServerBase* virtual_server_{nullptr};
            [[nodiscard]] ServerId get_server_id() const;
    };
}