#pragma once

#include "./VirtualServerBase.h"

namespace ts {
    namespace music {
        class MusicBotManager;
    }

    namespace stats {
        class ConnectionStatistics;
    }

    namespace server {
        namespace tokens {
            class TokenManager;
        }

        namespace bans {
            class BanManager;
        }

        namespace conversation {
            class ConversationManager;
        }
    }
}

namespace ts::server::vserver {
    class VirtualServer : public VirtualServerBase {
        public:
            bool initialize(std::string &string) override;

            /* managers */
            [[nodiscard]] inline tokens::TokenManager& token_manager() { return *this->token_manager_; }
            [[nodiscard]] inline bans::BanManager& ban_manager() { return *this->ban_manager_; }
            [[nodiscard]] inline const std::shared_ptr<music::MusicBotManager>& music_bot_manager() { return this->music_bots_; }
            [[nodiscard]] inline const std::shared_ptr<conversation::ConversationManager>& conversation_manager() { return this->conversation_manager_; }

        private:
            std::unique_ptr<tokens::TokenManager> token_manager_;
            std::unique_ptr<bans::BanManager> ban_manager_;

            std::shared_ptr<music::MusicBotManager> music_bots_;
            std::shared_ptr<conversation::ConversationManager> conversation_manager_;

        protected:
            VirtualServerStartResult start_server_virtual_(ts::rwshared_lock<ts::rw_mutex> &lock, std::string &string) override;
            VirtualServerStartResult start_server_(ts::rwshared_lock<ts::rw_mutex> &lock, std::string &string) override;
            void stop_server_(ts::rwshared_lock<ts::rw_mutex> &lock) override;
            void execute_tick(const std::chrono::system_clock::time_point &point) override;
    };
}